"""
Treat all arguments passed by the user.
"""

def treat_args(data, args) -> bool:
    '''
    Lower all user inputs an compare to lowered metadata. Evaluate the
    presence of an argument and include it in treatment

    Arguments:
    data -- MIBiG metadata from json files
    args -- object of class Args containing user inputs
    '''

    add: bool = True

    if args.organism:
        add &= (
            args.organism.lower() in data['taxonomy']['name'].lower()
            )
    if args.product:
        add &= args.product.lower() in [
            c.get('name').lower() for c in data['compounds']
            ]
    if args.biosynt:
        add &= args.biosynt.lower() in [
            b.get('class').lower() for b in data['biosynthesis']['classes']
            ]
    if args.completeness != 'all':
        add &= (
            data['completeness'].lower()
            == args.completeness.lower()
            )
    if args.quality != 'all':
        add &= (
            data['quality'].lower() == args.quality.lower()
            )

    return add



"""
The main module of the package.
Responsible for calling other modules' functions.
"""

import sys
import os
from src.pymibig.download_json import download_json
from src.pymibig.download_nucl import download_nucl
from src.pymibig.download_aa import download_aa
from src.pymibig.save_access_codes import save_access_codes
from src.pymibig.save_nucl import save_nucl
from src.pymibig.save_aa import save_aa
from src.pymibig.console import console
from src.pymibig.get_args import get_args

basedir: str = os.path.dirname(__file__)

def main() -> None:
    '''
    Execute functions calls, takes no arguments and has no return.
    '''
    args = get_args()

    download_json(basedir)
    download_nucl(basedir)
    download_aa(basedir)

    access_codes = save_access_codes(args, basedir)
    save_nucl(args, access_codes, basedir)
    save_aa(args, access_codes, basedir)

    console.print(
        '[bold blue]Task completed! Check your result files.[/bold blue]'
        )

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit('\nExecution interrupted by the user.')



"""
Save the matched nucleotide sequences in FASTa format.
"""

import sys
import tarfile
import io
from itertools import islice
from rich.progress import track
from Bio import SeqIO
from src.pymibig.console import console
from src.pymibig.constants import NUCLEOTIDE

def save_nucl(args, access_codes: str, basedir: str) -> None:
    '''
    Save the desired nucleotide sequences in a FASTa file.

    Arguments:
    args -- object of class Args containing user inputs
    access_codes -- codes list to retrieve from database
    basedir -- main module path
    '''
    desired_seqs: list = []

    try:
        with tarfile.open(f'{basedir}/src/db/{NUCLEOTIDE}') as tar:
            for member in track(islice(tar, 1, None),
            description='[bold green]Saving nucleotide '
                        'sequences...[/bold green]',
            total=len(tar.getmembers())-1):
                with tar.extractfile(member) as handle:
                    seq = SeqIO.read(
                        io.TextIOWrapper(handle),
                        'genbank')
                    if any(code in seq.id for code in access_codes):
                        desired_seqs.append(seq)

        SeqIO.write(
            desired_seqs,
            f'{args.create_prefix}_nucl.fasta', 'fasta')
    except PermissionError:
        console.print(
            '[bold red]Permission to read directory or write file '
            'denied.[/bold red]'
            )
        sys.exit()



"""
Download JSON metadta files from MIBiG
"""

import os
import sys
import requests
from rich.progress import track
from src.pymibig.console import console
from src.pymibig.constants import JSON_LINK, METADATA, CHUNK_SIZE

def download_json(basedir: str) -> None:
    '''
    Download JSON files tar.gz compressed.

    Argumnents:
    basedir -- main module path
    '''
    if not os.path.exists(f'{basedir}/src/db/{METADATA}'):
        try:
            resp = requests.get(JSON_LINK, stream=True, timeout=60)
            total_size = int(resp.headers.get('content-length', 0))
            with open(f'{basedir}/src/db/{METADATA}', mode='wb') as file:
                for chunk in track(resp.iter_content(chunk_size=CHUNK_SIZE),
                description='[bold green]Downloading MIBiG metadata'
                            '...[/bold green]',
                total=total_size / CHUNK_SIZE):
                    file.write(chunk)
        except PermissionError:
            console.print('[bold red]File can not be writen.[/bold red]')
            sys.exit()
        except requests.exceptions.RequestException:
            console.print('[bold red]Connection error or timed '
                          'out.[/bold red]')
    else:
        console.print('[bold green]Loading JSON metadata...[/bold green]')



"""
class Args definition
"""

class Args:
    '''
    Create a prefix for resulting filenames.
    '''
    def __init__(self, args):
        self.organism = args.organism
        self.product = args.product
        self.biosynt = args.biosynt
        self.completeness = args.completeness
        self.quality = args.quality
        self.prefix: str = ''

    @property
    def create_prefix(self) -> str:
        '''
        Take user arguments and compose filenames prefix with them.
        '''

        org = f'{self.organism}_' if self.organism else ""
        prod = f'{self.product}_' if self.product else ""
        bio = f'{self.biosynt}_' if self.biosynt else ""
        comp = f'{self.completeness}'
        quali = f'_{self.quality}'

        self.prefix = org + prod + bio + comp + quali

        return self.prefix



"""
Constant values
"""

JSON_LINK: str = "https://dl.secondarymetabolites.org/mibig/mibig_json_4.0.tar.gz"
GBK_LINK: str = "https://dl.secondarymetabolites.org/mibig/mibig_gbk_4.0.tar.gz"
FASTA_LINK: str = "https://dl.secondarymetabolites.org/mibig/mibig_prot_seqs_4.0.fasta"
METADATA: str = "mibig_json_4.0.tar.gz"
NUCLEOTIDE: str = "mibig_gbk_4.0.tar.gz"
AA: str = "mibig_prot_seqs_4.0.fasta.gz"
CHUNK_SIZE: int = 10 * 1024



"""
Save target's BGCs access codes.
"""

import sys
import json
import tarfile
from itertools import islice
import pandas as pd
from rich.progress import track
from src.pymibig.console import console
from src.pymibig.treat_args import treat_args
from src.pymibig.constants import METADATA

def save_access_codes(args, basedir) -> list:
    '''
    Create a txt file listing BGCs codes

    Arguments:
    args -- object of class Args containing user inputs
    basedir -- main module path
    '''
    df = pd.DataFrame(
        columns=['Code', 'Organism', 'Compounds', 'Biosynthetic Class',
                 'Completeness', 'Quality', 'Status']
        )

    try:
        with tarfile.open(f'{basedir}/src/db/{METADATA}') as tar:
            for member in track(islice(tar, 1, None),
            description='[bold green]Searching target access '
                        'codes...[/bold green]',
            total=len(tar.getmembers())-1):
                with tar.extractfile(member) as handle:
                    data = json.load(handle)
                if treat_args(data, args):
                    df.loc[member, 'Code'] = data['accession']
                    df.loc[member, 'Organism'] = data['taxonomy']['name']
                    df.loc[member, 'Compounds'] = ', '.join(
                        [c.get('name') for c in data['compounds']]
                        )
                    df.loc[member, 'Biosynthetic Class'] = ', '.join(
                        [b.get('class') for b in data['biosynthesis']['classes']]
                        )
                    df.loc[member, 'Completeness'] = data['completeness']
                    df.loc[member, 'Quality'] = data['quality']
                    df.loc[member, 'Status'] = data['status']
        if df.empty:
            console.print('[bold yellow]Your search had no '
                          'match[/bold yellow]')
            sys.exit()
        df.to_csv(
            f'{args.create_prefix}_codes.tsv', sep='\t', index=False
            )
        return df['Code'].to_list()
    except PermissionError:
        console.print(
            '[bold red]Permission to read directory or write '
            'file denied.[/bold red]'
            )
        sys.exit()
    except FileNotFoundError:
        console.print(f'[bold red]{tar.name} not found.[/bold red]')
        sys.exit()



"""
Save the matched amino acid sequences in FASTa format.
"""

import sys
import gzip
from rich.progress import track
from Bio import SeqIO
from src.pymibig.console import console
from src.pymibig.constants import AA

def save_aa(args, access_codes: str, basedir: str) -> None:
    '''
    Save the desired amino acid sequences in a FASTa file.

    Arguments:
    args -- object of class Args containing user inputs
    basedir -- main module path
    '''
    desired_seqs: list = []

    try:
        with gzip.open(f'{basedir}/src/db/{AA}', mode='rt') as gz:
            total = len(list(SeqIO.parse(gz, 'fasta')))

        with gzip.open(f'{basedir}/src/db/{AA}', mode='rt') as gz:
            for seq in track(SeqIO.parse(gz, 'fasta'),
            description='[bold green]Saving amino acid '
                        'sequences...[/bold green]',
            total=total):
                if any(code in seq.id for code in access_codes):
                    desired_seqs.append(seq)

        SeqIO.write(
            desired_seqs,
            f'{args.create_prefix}_aa.fasta', 'fasta')
    except PermissionError:
        console.print(
            '[bold red]Permission to read directory or write file '
            'denied.[/bold red]'
            )
        sys.exit()



"""
Get arguments from user.
"""

import argparse
from src.pymibig.args_class import Args


def get_args() -> Args:
    '''
    Recieve user arguments and return them to main function.
    '''
    parser = argparse.ArgumentParser(
        prog='pyMIBiG',
        description='A small tool to download, match and save targeted '
                    'sequences from MIBiG.'
    )
    parser.add_argument('-o', '--organism',
                        help='Organism name to query in database.',
                        type=str)
    parser.add_argument('-p', '--product', help='Compound to query in database.',
                        type=str)
    parser.add_argument('-b', '--biosynt',
                        help='Biosynthetic class to query in database.',
                        type=str)
    parser.add_argument('-c', '--completeness', help='Loci completeness.',
        type=str, choices=['complete', 'incomplete', 'unknown', 'all'],
        default='all')
    parser.add_argument('-q', '--quality', help='Entry quality level.',
        type=str, choices=['low', 'medium', 'high', 'questionable', 'all'],
        default='all')

    args = parser.parse_args()

    if not (args.organism or args.product or args.biosynt):
        parser.error('At least one of the following arguments is requires: '
                     'organism, product, biosynt')

    return Args(args)
